@@ -44,7 +44,7 @@
 #  define PRId64 "I64d"
 #endif
 
-#include <oggz/oggz.h>
+#include "oggz/oggz.h"
 
 #include "oggz_tools.h"
 
@@ -354,9 +354,10 @@ read_page (OGGZ * oggz, const ogg_page * og, long serialno, void * user_data)
 }
 
 static int
-read_packet (OGGZ * oggz, ogg_packet * op, long serialno, void * user_data)
+read_packet (OGGZ * oggz, oggz_packet * zp, long serialno, void * user_data)
 {
   OVData * ovdata = (OVData *)user_data;
+  ogg_packet * op = &zp->op;
   timestamp_t timestamp;
   int flush;
   int ret = 0, feed_err = 0, i;
